(*
================================================================================
TPGTrayIcon95 - Super TrayIcon VCL
================================================================================
  Note: See the help file PGTRAY95.HLP for information concerning published
  properties and methods.
                         
  Classes:
  ==============================================================================
    Name                      Purpose
    TPGTrayIcon95             Windows 95 TrayIcon Component.

    TPGTrayIconAnimation      Used to implement the Animation property of the
                              TPGTrayIcon95 component.

    TPGTrayIconMsgWindow      Used to intercept messages sent to the TrayIcon
                              by Windows, and more generally moves the majority
                              of the mouse handling routines out of the
                              Super TrayIcon VCL.

    TPGTimerManager           Manages a group of timed events using a single
                              Windows Timer.  Used to conserve windows timer
                              resources for 16 bit applications.

  Associated Files:
  ==============================================================================
    PGBASE.INC                An include unit that supplies some general
                              conditional defines to make it easier to figure
                              out which code is associated with the different
                              versions of Delphi and C++ Builder respectively.

    CALL32NT.INC              An include unit that provides access to the
                              Win32 API for Delphi 1.0 applications. Note
                              that this also involves some code in the
                              initialization section.  Used by the Delphi 1.0
                              version of the component.  Many thanks to
                              Christian Ghisler, CIS: 100332,1175 for releasing
                              CALL32NT.PAS into the public domain, and for
                              helping me figure out how to convert a 16 bit
                              HIcon to a 32 bit HIcon.

    PGAPPWND.PAS              Used to implement a hidden application window
                              class. Used by the Delphi 1.0 Solution to solve
                              a Windows NT 4.X LoadIcon problem, and by the
                              Delphi 2.0 solution to initially hide the
                              application on initial startup.

    PGHIDE32.PAS              Used exclusively by the Delphi 2.0 solution in
                              conjunction with the PGAPPWND.PAS file to
                              initially hide the application on initial startup.
                              
    PGTRAY16.DCR              Component Resource for Delphi 1.0
    PGTRAY32.DCR              Component Resource for Delphi 2.0

  History:
  ==============================================================================
    Ver  Date(d/m/y)  Programmer   Comments
    1.0  06/03/96     No One   Creation.

    2.0  20/05/96     No One   Added the State and Animation properties, as
                                   well as the ability to use the component
                                   in Delphi 1.0 and Delphi 2.0.

    2.1  18/06/96     No One   Exposed a method to display the popup menu at
                                   the cursor position, "ShowPopupAtCursor".

    2.2  26/08/96     No One   Fixed a bug in the Delphi 2 version of the
                                   component that stopped the reception of tray
                                   messages.  Also introduced the "MouseOver"
                                   event, the "ShowApplication" and
                                   "HideApplication" methods, and made the
                                   popup menu "non sticky"

    3.0  08/02/97     No One   Modified the "TPGTrayIconMsgWindow" class.
                                   It is now derived from TComponent instead of
                                   TWinControl to ensure that the callback
                                   window handle that Windows sends our TrayIcon
                                   messages to remains constant.
                                   It would change when the Form that the
                                   TrayIcon is on was Destroyed and Recreated,
                                   as modifying properties like TForm.FormStyle
                                   do when they are changed.
                                   - Introduced the Options property, the
                                   OnMouseEnter, OnMouseExit, OnAppMinimize, and
                                   OnAppRestore events.
                                   - Fixed 16 bit Win NT 4.0 bug that stopped
                                   the TrayIcon from being seen in the Tray.
                                   - Introduced more thorough solutions to hide
                                   the application on startup for Delphi 1.0 and
                                   2.0.  In conjunction with this, revised the
                                   ShowApplication and HideApplication methods
                                   to work when more than one form is visible
                                   on the screen.

    3.1  16/07/97     No One   - Exposed the PostEvent method, and made
                                   the ShowApplication, HideApplication and
                                   Click events use it to execute their code
                                   asynchronously.  This was done to avoid
                                   bugs that occur with the delay between
                                   operation requests and application responses.
                                   - Fixed a bug that would cause Delphi itself
                                   to hide it's taskbar icon when minimized at
                                   design time.

    3.2 23/07/97      No One   - Added the ShowForm and ShowModalForm
                                   methods to allow Delphi 3.0 users to Show
                                   and Hide forms without the Applications'
                                   TaskBar icon being redisplayed.  This problem
                                   was introduced in Delphi 3.0 by a change that
                                   Borland made to the Forms.pas file (see the
                                   TApplication.UpdateVisible routine).

    3.3 23/07/97      No One   - Updated the ShowForm, ShowModalForm,
                                   ShowApplication, HideApplication to allow
                                   forms and applications to be shown and hidden
                                   without displaying the Applications's TaskBar
                                   icon ever.  This also involved some additions
                                   to the Options property, toCloseFormsOnHideApp,
                                   and toHideTaskBarIconOnShowApp.
                                   - As suggested by Richard K Peterson of
                                   Common Sense Solutions{css}, turned the Icon
                                   Index property into a Read AND write property,
                                   so the default animation sequence can be
                                   overriden.
================================================================================
Copyright  1996,1997 No Name Software.  All Rights Reserved
================================================================================
*)

unit PGTray95;

interface

{ Include some general compiler switches. }
{$I PGBASE.INC}

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, ExtCtrls, Dialogs, IniFiles, ShellApi, Menus, PGAppWnd;

const
  { Callback message and id used by Windows to send messages to the TrayIcon. }
  WM_TRAY_CALLBACK = WM_USER + 100;
  WM_PGPOSTEVENT = WM_USER + 101;
  TRAY_ICON_ID = 42;

  { Port of the TrayIcon messages to 16 bit windows.  These values }
  { are identical to the values provided by the ShellApi unit that }
  { comes with Delphi 2.0, and are used when sending messages in both the }
  { Delphi 1.0 and 2.0 versions of the component. }

  PGNIM_ADD     = $00000000;  { Used to add the icon to the tray. }
  PGNIM_MODIFY  = $00000001;  { Used to modify the icon in the tray. }
  PGNIM_DELETE  = $00000002;  { Used to remove the icon from the tray. }

  PGNIF_MESSAGE = $00000001;  { Flags that indicate which fields in the }
  PGNIF_ICON    = $00000002;  { PGNOTIFYICONDATA data structure contain }
  PGNIF_TIP     = $00000004;  { valid information.  These flags can be }
                              { used in combination with each other. }

  { Miscellaneous Constants }
  NID_MAXTIP         = 63;    { The maximum length of the icon's hint. }
  Animation_Interval = 1000;  { The default time between animation frames. }

type
  { TPGTrayIconAnimation - Used to implement the Animation property of }
  {   the TPGTrayIcon95 component.  Note that this component is derived from }
  {   TPersistent, because anything derived directly from TPersistent can be }
  {   stored in a form file, and is automatically published in the object }
  {   inspector.  This avoided having to create a property editor for the }
  {   component.  Note that the component is initialized when it is created }
  {   in the constructor of the TPGTrayIcon95 component, since you can't }
  {   override the constructor of TPersistent objects. }
  TPGTrayIcon95 = class;
  TPGTimerItem = class;
  TPGTimerManager = class;

  TPGTrayIconAnimationOptions = (aoIconsInExe, aoIconsAreFiles);
  TPGTrayIconAnimation = class(TPersistent)
  protected
    AnimationTimer: TPGTimerItem;
    TrayIcon: TPGTrayIcon95;
    FIconIndex: Integer;
    FInterval: Integer;
    FIcons: TStringList;
    FOptions: TPGTrayIconAnimationOptions;
    procedure SetInterval(Value: Integer);
    procedure SetIcons(Value: TStringList);
    procedure SetIconIndex(Value: Integer);                                     {css}
    procedure StartAnimation;
    procedure StopAnimation;
    procedure AnimationEvent(Sender: TObject);
  public
    property IconIndex: Integer read FIconIndex write SetIconIndex;             {css}
  published
    property Interval: Integer read FInterval write SetInterval default Animation_Interval;
    property Icons: TStringList read FIcons write SetIcons;
    property Options: TPGTrayIconAnimationOptions read FOptions write FOptions default aoIconsAreFiles;
  end;

  { Port of the 32 bit TrayNotification data structure to Delphi 1.0.  }
  { Note: This ported structure is used for both the Delphi 1.0 and 2.0 }
  { versions of the component.  The Delphi 2.0 version converts all }
  { values stored in this structure into the ShellApi unit version of }
  { the structure just before it uses the structure in the "NotifyShell" }
  { method of the TPGTrayIcon95 component. }
  PPGNOTIFYICONDATA = ^PGNOTIFYICONDATA;
  PGNOTIFYICONDATA = record
    cbSize: LongInt;
    Wnd: LongInt;
    uID: LongInt;
    uFlags: LongInt;
    uCallbackMessage: LongInt;
    hIcon: LongInt;
    szTip: array[0..63] of char;
  end;

  TPGTrayIconMsgWindow = class;
  TPGTrayIconState = (tsEnabled, tsDisabled, tsAnimated);
  TPGTrayIconOption = (toMinimizeToTray, toUnclutterDelphi1TaskBar, toCloseFormsOnHideApp, toHideTaskBarIconOnShowApp);
  TPGTrayIconOptions = set of TPGTrayIconOption;
  TPGTrayIconStateChangingEvent = procedure (Sender: TObject; var NewState: TPGTrayIconState) of object;

  { TPGTrayIcon95 - Windows 95/NT TrayIcon Component }
  {   Note: "LoadCurrentIcon" and "NotifyShell" are the key methods here. }
  TPGTrayIcon95 = class(TComponent)
  protected
    FAnimation: TPGTrayIconAnimation;
    FActive: Boolean;
    FAutoPopup: Boolean;
    FDisabledIcon: TIcon;
    FForm: TForm;
    FHandle: THandle;
    FHint: string;
    FIcon: TIcon;
    FState: TPGTrayIconState;
    FOptions: TPGTrayIconOptions;
    FOnAnimation: TNotifyEvent;
    FOnClick: TNotifyEvent;
    FOnDblClick: TNotifyEvent;
    FOnMouseMove: TMouseMoveEvent;
    FOnMouseEnter: TMouseMoveEvent;
    FOnMouseExit: TMouseMoveEvent;
    FOnStateChanging: TPGTrayIconStateChangingEvent;
    FOnAppMinimize: TNotifyEvent;
    FOnAppRestore: TNotifyEvent;
    FPopupMenu: TPopupMenu;
    FShowHint: Boolean;
    FTimerManager: TPGTimerManager;
    CurrentIcon: TIcon;
    tnd: PGNOTIFYICONDATA;
    PostedEvent: TNotifyEvent;
    PostedEventSender: TObject;
    SaveOnShowEvent: TNotifyEvent;
    AlreadyShowingForm: Boolean;
    ApplicationWasHidden: Boolean;
    MessageWindow: TPGTrayIconMsgWindow;
    function GetHint: string;
    function GetHandle: THandle;
    function GetClickedMouseButton: TMouseButton;
    procedure SetActive(AnActive: Boolean);
    procedure SetDisabledIcon(Value: TIcon);
    procedure SetHint(Value: string);
    procedure SetIcon(Value: TIcon);
    procedure SetOptions(Value: TPGTrayIconOptions);
    procedure SetShowHint(Value: Boolean);
    procedure SetState(AState: TPGTrayIconState);
    procedure FireEvent(AnEvent:TNotifyEvent);
    procedure IconChanged(Sender: TObject);
    procedure LoadAnimationIcon(const AnIconIndex: Integer);
    procedure LoadCurrentIcon;
    procedure Loaded; override;
    procedure ApplicationMinimize(Sender: TObject);
    procedure ApplicationRestore(Sender: TObject);
    function DoShowForm(const AForm: TForm; const ShowModal: Boolean): Integer;
    function HIcon32: LongInt;
    function ApplicationIsHidden: Boolean;
    procedure DoShowApplication(Sender: TObject);
    procedure DoHideApplication(Sender: TObject);
    procedure DoFormOnShow(Sender: TObject);
    function NotifyOnFormClose(AForm: TForm): Boolean;
    procedure FormOnCloseNotification(Sender: TObject);
    procedure DoHideAppTaskBarIcon(Sender: TObject);
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    function ShouldDoMouseMove: Boolean;
    function ShouldDoMouseEnterOrMouseExit: Boolean;
    function ShouldDisplayPopup: Boolean;
  public
    procedure ShowIcon;
    procedure HideIcon;
    procedure ShowApplication; virtual;
    procedure HideApplication; virtual;
    procedure ShowForm(const AForm: TForm);
    function ShowModalForm(const AForm: TForm): Integer;
    procedure CloseForm(const AForm: TForm);
    procedure LockScreenUpdates(ShouldLock: Boolean); virtual;
    procedure LockTaskBarUpdates(ShouldLock: Boolean);
    procedure Refresh;
    procedure RefreshTaskList;
    procedure ShowPopupAtCursor;
    procedure NotifyShell(dwMessage: LongInt);
    procedure PostEvent(AnEvent: TNotifyEvent);
    procedure PostEventFromObject(AnEvent: TNotifyEvent; ASender: TObject);
    procedure DoSetForegroundWindow(Sender: TObject);
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  public
    property ClickedMouseButton: TMouseButton read GetClickedMouseButton;
    property Form: TForm read FForm;
    property Handle: THandle read GetHandle;
    property TimerManager: TPGTimerManager read FTimerManager;
  published
    { Store both icon properties before the Active property.  If they aren't a }
    { blank icon will be displayed if the component is "Active" on FormLoad. }
    property Icon: TIcon read FIcon write SetIcon;
    property DisabledIcon: TIcon read FDisabledIcon write SetDisabledIcon;
    property Animation: TPGTrayIconAnimation read FAnimation write FAnimation;
    property Active: Boolean read FActive write SetActive default false;
    property AutoPopup: Boolean read FAutoPopup write FAutoPopup default false;
    property Hint: string read FHint write SetHint;
    property OnAnimation: TNotifyEvent read FOnAnimation write FOnAnimation;
    property OnAppMinimize: TNotifyEvent read FOnAppMinimize write FOnAppMinimize;
    property OnAppRestore: TNotifyEvent read FOnAppRestore write FOnAppRestore;
    property OnClick: TNotifyEvent read FOnClick write FOnClick;
    property OnDblClick: TNotifyEvent read FOnDblClick write FOnDblClick;
    property OnMouseMove: TMouseMoveEvent read FOnMouseMove write FOnMouseMove;
    property OnMouseEnter: TMouseMoveEvent read FOnMouseEnter write FOnMouseEnter;
    property OnMouseExit: TMouseMoveEvent read FOnMouseExit write FOnMouseExit;
    property OnStateChanging: TPGTrayIconStateChangingEvent read FOnStateChanging write FOnStateChanging;
    property Options: TPGTrayIconOptions read FOptions write SetOptions
                      default [toMinimizeToTray, toUnclutterDelphi1TaskBar, toCloseFormsOnHideApp];
    property PopupMenu: TPopupMenu read FPopupMenu write FPopupMenu;
    property ShowHint: Boolean read FShowHint write SetShowHint default true;
    property State: TPGTrayIconState read FState write SetState default tsEnabled;
  end;

  { TPGTrayIconMsgWindow - Message Window component }
  {   Receives messages sent to the TPGTrayIcon95 by Windows, and more }
  {   generally moves the majority of the mouse handling routines out }
  {   of the TPGTrayIcon95 component. }
  TPGTimerInterval = 55..100000;
  TPGTrayIconMsgWindow = class(TComponent)
  protected
    ClickedMouseButton: TMouseButton;
    TrayIcon: TPGTrayIcon95;
    DoubleClickHasOccurred: Boolean;
    WaitForClickEventTimer: TPGTimerItem;
    MouseMoveTimer: TPGTimerItem;
    ptMousePosition: TPoint;
    AlreadyFiredMouseEnterEvent: Boolean;
    HitTrayIcon: Boolean;
    procedure TrayIconCallback(var Message: TMessage); message WM_TRAY_CALLBACK;
    procedure StartWaitingForDoubleClick;
    procedure DoClickEventAsynchronously(AnInterval: TPGTimerInterval);
    procedure DoneWaitingForDoubleClick(Sender: TObject);
    procedure StartTrackingMouseMovement(ptCursor: TPoint);
    procedure StopTrackingMouseMovement;
    procedure TrackMouseMovement(Sender: TObject);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  end;

  { TPGTimerManager - Uses a single windows timer to manage a group of timed }
  { events.  Used to conserve windows timer resources for 16 bit applications. }
  TPGTimerManager = class(TComponent)
  protected
    FEnabled: Boolean;
    TimerItems: TList;
    procedure SetEnabled(Value: Boolean);
    procedure DoTimeouts(Sender: TObject);
  public
    Timer: TTimer;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function AddTimer(AnEnabled: Boolean; AnInterval: TPGTimerInterval; AnEvent: TNotifyEvent): TPGTimerItem;
    procedure DeleteTimer(AnId: Integer);
    function Timers(AnIndex: Integer): TPGTimerItem;
    function TimerCount: Integer;
    procedure RefreshInterval;
  published    
    property Enabled: Boolean read FEnabled write SetEnabled;
  end;

  TPGTimerItem = class(TComponent)
  protected
    TimerManager: TPGTimerManager;
    FInterval: TPGTimerInterval;
    FEnabled: Boolean;
    OnTimer: TNotifyEvent;
  protected
    procedure SetEnabled(Value: Boolean);
    procedure SetInterval(Value: TPGTimerInterval);
  public
    LastTickCount: Double;
    Tag: LongInt;
    constructor Create(AOwner: TComponent); override;
    function ShouldTimeout: Boolean;
    procedure TimeOut;
  published
    property Enabled: Boolean read FEnabled write SetEnabled;
    property Interval: TPGTimerInterval read FInterval write SetInterval;
  end;

{ Used to hook into a forms OnClose event so we can be notified just }
{ before a Form closes.  Used to automatically hide the Applications }
{ TaskBar icon on closing a form in Delphi 3.X }
TPGNotifyOnFormClose = class(TComponent)
public
  Form: TForm;
  FormOnClose: TCloseEvent;
  NotifyOnFormCloseEvent: TNotifyEvent;
  procedure AddCloseEvent(AForm: TForm; AnEvent: TNotifyEvent);
  procedure DoFormOnClose(Sender: TObject; var Action: TCloseAction);
  destructor Destroy; override;
end;

{ Replaces the Application.Run line for Delphi 1.0 and 2.0. }
procedure RunTrayApplication;

{ Returns the State keys that are currently depressed on the keyboard.  Used }
{ with the mouse events. }
function PGShiftState: TShiftState;

procedure Register;

var
  { The time to wait between a single click and a double click when }
  { processing mouse events. }
  PGTrayIconDoubleClickTime: TPGTimerInterval;

{ Delphi 1.0 specific code - Declare Win32 functions under Windows 3.1 }
{ These functions are declared using the CALL32NT.INC interface, and }
{ are necessary to properly implement a TrayIcon under Win95 using Delphi 1.0 }
{$IFNDEF WIN32}
  Shell_NotifyIcon: function (dwMessage: LongInt; lpData:PPGNOTIFYICONDATA; id:LongInt): LongBool;
  idShell_NotifyIcon: LongInt;

  PGWowHandle32: function(dwHandle,dwType,id:longint):longint;
  idPGWowHandle32: LongInt;

  PGGetClassLong: function(hWnd: LongInt; AnIndex: LongInt; id: LongInt): LongInt;
  idPGGetClassLong: LongInt;

  SetForegroundWindow: function (hWnd: LongInt; id: LongInt): LongInt;
  idSetForegroundWindow: LongInt;
{$ENDIF}

implementation

{$IFDEF PGShareware_Version}
uses
  StdCtrls;
{$ENDIF}

{$IFNDEF WIN32}
{ Returns whether or not we are currently running under Windows NT }
function RunningWinNT: Boolean;
begin
  Result := (GetWinFlags and $4000 <> 0);
end;
{$ENDIF}


{ Delphi 1.0 specific code - Provides Delphi 1.0 access to the Win32 API. }
{ Include a modified version of CALL32NT.PAS in this component.  This }
{ allows us to include the functionality of this unit without having to }
{ send the CALL32NT.DCU file along with the component. }
{$IFNDEF WIN32}
  {$I CALL32NT.INC}
{$ENDIF}

{ Include the components palette bitmap as a resource.  This allows }
{ us to easily side step the incompatability of 16 and 32 bit resources. }
{$IFDEF WIN32}
  {$R PGTRAY32.DCR}
{$ELSE}
  {$R PGTRAY16.DCR}
{$ENDIF}

{ Include the registration dialog if this is the shareware version. }
{$IFDEF PGShareware_Version}

procedure PGAssignSharewareInfo(Info: TStrings);
begin
  Info.Add('Super TrayIcon VCL v3.3 for Delphi 1.0, 2.0, 3.0');
  Info.Add('and C++ Builder 1.0');
  Info.Add('');
  Info.Add('Adds animated icons to the Windows 95 / Windows NT ' +
           'system tray.  Register the component for 10$ US or ' +
           'obtain the complete source code for only 25$ US.');
  Info.Add('');
end;
  {$I PGSHARE.INC}
{$ENDIF}

procedure Register;
begin
  RegisterComponents('NoName', [TPGTrayIcon95]);
end;

{ Used to determine the state of the Alt, Ctrl, and Shift keys, and }
{ the mouse buttons.  Used by some of the mouse events.}
function PGShiftState: TShiftState;
begin
  Result := [];
  if GetKeyState(vk_Shift) < 0 then
    Include(Result, ssShift);
  if GetKeyState(vk_Control) < 0 then
    Include(Result, ssCtrl);
  if GetKeyState(vk_Menu) < 0 then
    Include(Result, ssAlt);
  if GetKeyState(vk_LButton) < 0 then
    Include(Result, ssLeft);
  if GetKeyState(vk_MButton) < 0 then
    Include(Result, ssMiddle);
  if GetKeyState(vk_RButton) < 0 then
    Include(Result, ssRight);
end;

{ SetForegroundWindow API wrapper. }
procedure PGSetForegroundWindow(AHandle: LongInt);
begin
{$IFDEF WIN32}
  SetForegroundWindow(AHandle);
{$ELSE}
  SetForegroundWindow(AHandle, idSetForegroundWindow);
{$ENDIF}
end;

{$IFDEF WIN32}
function PGGetAnimation: Boolean;
var
  Info: TAnimationInfo;
begin
  Info.cbSize := SizeOf(TAnimationInfo);
//  AResult := (Info.iMinAnimate <> 0);
  if SystemParametersInfo(SPI_GETANIMATION, SizeOf(Info), @Info, 0) then
    Result := Boolean(Info.iMinAnimate)
   else
    Result := False;
end;

procedure PGSetAnimation(Value: Boolean);
var
  Info: TAnimationInfo;
begin
  Info.cbSize := SizeOf(TAnimationInfo);
  BOOL(Info.iMinAnimate) := Value;
  SystemParametersInfo(SPI_SETANIMATION, SizeOf(Info), @Info, 0);
end;
{$ENDIF}

procedure PGShowWinNoAnimate(Handle: HWnd; CmdShow: Integer);
var
  Animation: Boolean;
begin
{$IFDEF WIN32}
  Animation := PGGetAnimation;
  if Animation then PGSetAnimation(False);
  ShowWindow(Handle, CmdShow);
  if Animation then PGSetAnimation(True);
{$ELSE}
  ShowWindow(Handle, CmdShow);
{$ENDIF}
end;

{ TPGTrayIconAnimation }
procedure TPGTrayIconAnimation.StartAnimation;
begin
  { Stop any previous animations that may have been going on. }
  StopAnimation;

  { Make the animation event happen as soon as the animated }
  { icon is shown in the system tray. }
  FIconIndex := 0;
  AnimationEvent(Self);

  { A developer could have used the OnAnimation event to stop the animation }
  { from happening.  If they did, stop preparing for the animation to happen }
  { and exit the procedure. }
  if (TrayIcon.FState <> tsAnimated) then
    exit;

  { Create the animation timer that is used to track when different frames }
  { of the animation should be displayed. }
  AnimationTimer.Interval := FInterval;
  AnimationTimer.Enabled := true;
end;

procedure TPGTrayIconAnimation.StopAnimation;
begin
  AnimationTimer.Enabled := false;
end;

{ Loads the next icon in the animation, ignoring icons that }
{ can't be loaded. Also fires the OnAnimation event. }
procedure TPGTrayIconAnimation.AnimationEvent(Sender: TObject);
begin
  try
    TrayIcon.FireEvent(TrayIcon.FOnAnimation);
    TrayIcon.LoadCurrentIcon;
  except
    on EFOpenError do;
    else
      raise;
  end;

  { Increment to the next icon in the list. }
  if FIconIndex < (FIcons.Count - 1) then
    Inc(FIconIndex)
  else
    FIconIndex := 0;
end;

{ Used to set the speed of a TrayIcon animation. }
procedure TPGTrayIconAnimation.SetInterval(Value: Integer);
begin
  if FInterval = Value then
    exit;

  FInterval := Value;

  { Restart the animation if there was one already in progress. }
  if (TrayIcon.State = tsAnimated) and TrayIcon.FActive then
    StartAnimation;
end;

procedure TPGTrayIconAnimation.SetIcons(Value: TStringList);
begin
  FIcons.Assign(Value);

  { Restart the animation if there was one already in progress. }
  if (TrayIcon.State = tsAnimated) and TrayIcon.FActive then
    FIconIndex := 0;
end;

procedure TPGTrayIconAnimation.SetIconIndex(Value: Integer);                    {css}
begin
  if (Value >= 0) and (Value < FIcons.Count) then
    begin
    FIconIndex:= Value;
    TrayIcon.LoadCurrentIcon;
    end;
end;


{ TPGTrayIconMsgWindow }
constructor TPGTrayIconMsgWindow.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  TrayIcon := TPGTrayIcon95(AOwner);
  AlreadyFiredMouseEnterEvent := false;
  HitTrayIcon := false;

  { Initialize the timer that is used to determine whether or not a }
  { TrayIcon has been clicked or double clicked. }
  WaitForClickEventTimer := TrayIcon.TimerManager.AddTimer(false, PGTrayIconDoubleClickTime, DoneWaitingForDoubleClick);

  { Initialize the timer that is used to determine whether or not }
  { OnMouseEnter and OnMouseExit mouse events have occurred. }
  MouseMoveTimer := TrayIcon.TimerManager.AddTimer(false, 70, TrackMouseMovement);
end;

destructor TPGTrayIconMsgWindow.Destroy;
begin
  inherited Destroy;
end;

{ This windows procedure is called by Windows whenever it has a message }
{ it wants to send to our component.  The procedure is coded to }
{ capture mouse messages, and to interpret from the messages whether or }
{ not the user clicked, or double clicked the TrayIcon.  }

{ Note: Whenever a mouse down message is received, the component starts a }
{ timer, and waits (the double click time in Windows) until it can }
{ determine if a single click or double click has occurred.  Once }
{ it has been figured out, the appropriate event is fired. }
procedure TPGTrayIconMsgWindow.TrayIconCallback(var Message : TMessage);
var
  AnEvent: TNotifyEvent;

  procedure ProcessTrayMessage;
  begin
    case Message.lParam of
      WM_LBUTTONDOWN, WM_MBUTTONDOWN, WM_RBUTTONDOWN:
        begin
        { Remember the  mouse button that was used. }
        if Message.lParam = WM_LBUTTONDOWN then
          ClickedMouseButton := mbLeft
        else if Message.lParam = WM_RBUTTONDOWN then
          ClickedMouseButton := mbRight
        else
          ClickedMouseButton := mbMiddle;

        { Display the popup menu if we're supposed to. }
        if (ClickedMouseButton = mbRight) and TrayIcon.ShouldDisplayPopup then
          begin
          TrayIcon.ShowPopupAtCursor;
          exit;
          end;

        { If events are defined for the Click and DoubleClick events, }
        { the only way we can figure out which one the user is doing }
        { is to wait around and see if it was a double click, so we'll }
        { start waiting. Note that the click and double click events are }
        { not immediately fired right now.  This is done to avoid clashes }
        { between outgoing (whatever you'd do in the Click/DblClick events) and }
        { incoming (the event that's currently being processed) processes. }
        { This isn't just "playing it safe".  It was a problem with creating }
        { an instance of an OLE automation server. }
        if Assigned(TrayIcon.FOnDblClick) then
          StartWaitingForDoubleClick
        else
          TrayIcon.PostEvent(DoneWaitingForDoubleClick);
        end;

      WM_LBUTTONDBLCLK, WM_MBUTTONDBLCLK, WM_RBUTTONDBLCLK:
        begin
        { Register that a double click has occurred.  This will be used in the }
        { timer event "DoneWaitingForDoubleClick" to determine which event should }
        { be fired, a single click or a double click. }
        DoubleClickHasOccurred := true;
        DoneWaitingForDoubleClick(Self);
        end;

      WM_MOUSEMOVE:
        begin
        GetCursorPos(ptMousePosition);

        { Fire OnMouseEnter event if it is defined.  Note that this starts }
        { a timer that is used to evaluate whether or not we have left }
        { the boundaries of the TrayIcon, and at that point fire the }
        { OnMouseExit event. }
        if TrayIcon.ShouldDoMouseEnterOrMouseExit then
          begin
          if not AlreadyFiredMouseEnterEvent then
            begin
            StartTrackingMouseMovement(ptMousePosition);
            exit;
            end
          end
        else
          AlreadyFiredMouseEnterEvent := true;

        { Fire OnMouseMove event if one is defined. }
        if TrayIcon.ShouldDoMouseMove then
          TrayIcon.FOnMouseMove(Self, PGShiftState, ptMousePosition.X, ptMousePosition.Y);
        end;
      end;
  end;
begin
  case Message.Msg of
    WM_QUERYENDSESSION:
      begin
      Message.Result := 1;
      end;
    WM_ENDSESSION:
      begin
      if Boolean(Message.wParam) then
        TrayIcon.HideIcon;
      end;
    WM_TRAY_CALLBACK:
      begin
      ProcessTrayMessage;
      end;
    WM_PGPOSTEVENT:
      begin
      if Assigned(TrayIcon.PostedEvent) then
        begin
        AnEvent := TrayIcon.PostedEvent;
        TrayIcon.PostedEvent := nil;
        try
          AnEvent(TrayIcon.PostedEventSender);
        except
          TrayIcon.PostedEvent := nil;
          TrayIcon.PostedEventSender := nil;
        end;
        end;
      end;
    end;
end;

{ Fire's the OnMouseEnter event, and starts a timer that is used to }
{ eventually fire the OnMouseExit event. }
procedure TPGTrayIconMsgWindow.StartTrackingMouseMovement(ptCursor: TPoint);
begin
  AlreadyFiredMouseEnterEvent := true;

  { Fire OnMouseEnter event. }
  MouseMoveTimer.Enabled := true;
  if Assigned(TrayIcon.FOnMouseEnter) then
    TrayIcon.FOnMouseEnter(TrayIcon, PGShiftState, ptCursor.X, ptCursor.Y);
end;

procedure TPGTrayIconMsgWindow.StopTrackingMouseMovement;
begin
  AlreadyFiredMouseEnterEvent := false;

  { Fire OnMouseExit event. }
  if Assigned(TrayIcon.FOnMouseExit) then
    TrayIcon.FOnMouseExit(TrayIcon, PGShiftState, ptMousePosition.X, ptMousePosition.Y);
  MouseMoveTimer.Enabled := false;
end;

{ Checks to see if the cursor is still in the bounds of the TrayIcon.  If }
{ it isn't, then the OnMouseExit event is fired.  This procedure is the }
{ OnTimer event for the MouseMoveTimer, and is enabled when the cursor }
{ enters the bounds of the TrayIcon. }
procedure TPGTrayIconMsgWindow.TrackMouseMovement(Sender: TObject);
var
  ptCursor: TPoint;
begin
  try
    { Find out where the cursor currently is, and compare that to where it }
    { was the last time that this event was run.  If they aren't relatively }
    { close, then this means that the cursor is no longer over the TrayIcon, }
    { and the OnMouseExit event should be fired. }
    GetCursorPos(ptCursor);
    if (Abs(ptMousePosition.X - ptCursor.X) > 2) or (Abs(ptMousePosition.Y - ptCursor.Y) > 2) then
      StopTrackingMouseMovement;
  except
  end;
end;

{ Starts a timer used in checking if a double click has occurred. }
procedure TPGTrayIconMsgWindow.StartWaitingForDoubleClick;
begin
  DoClickEventAsynchronously(PGTrayIconDoubleClickTime);
end;

{ Starts a timer used to "delay" processing click and double click events }
{ inside the main message loop.  This avoids the following OLE Automation }
{ Instantiation error: Error Code: RPC_E_CANTCALLOUT_ININPUTSYNCCALL - 'An }
{ outgoing call cannot be made since the application is dispatching an }
{ input-synchronous call'.  More information on this topic can be found at }
{ the following URL: http://www.microsoft.com/kb/articles/q131/0/56.htm }
procedure TPGTrayIconMsgWindow.DoClickEventAsynchronously(AnInterval: TPGTimerInterval);
begin
  DoubleClickHasOccurred := false;

  WaitForClickEventTimer.Interval := AnInterval;
  WaitForClickEventTimer.Enabled := true;
end;

{ Timer event that evaluates which mouse event has occurred, a single }
{ click, or a double click.  Once it has been determined, the appropriate }
{ event is fired. }
procedure TPGTrayIconMsgWindow.DoneWaitingForDoubleClick(Sender: TObject);
begin
  WaitForClickEventTimer.Enabled := false;
  if DoubleClickHasOccurred then
    TrayIcon.FireEvent(TrayIcon.FOnDblClick)
  else
    TrayIcon.FireEvent(TrayIcon.FOnClick);
end;


{ TPGTrayIcon95 }
constructor TPGTrayIcon95.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  { Initialize the Form property. }
  if not (AOwner is TForm) then
    raise EInvalidOperation.Create('TrayIcons must be owned by a Form !');
  FForm := TForm(AOwner);

  if PGShouldHideApplicationWindow and not (csDesigning in ComponentState) then
    begin
    ShowWindow(Application.Handle, sw_Hide);
    WinProcs.SetParent(Application.Handle, 0);
    end;

  { Create the TimerManager.  It is used to manage the Animation, Mouse Click }
  { and Mouse Enter/Leave events. }
  FTimerManager := TPGTimerManager.Create(Self);
  TimerManager.Enabled := true;

  { Create the message window that will receive tray icon messages }
  { sent to us by Windows. }
  MessageWindow := TPGTrayIconMsgWindow.Create(Self);
  FHandle := AllocateHWnd(MessageWindow.TrayIconCallback);

  { Initialize the tray notification data structure that will be used when }
  { sending Windows messages about our tray icon using the "NotifyShell" method. }
  tnd.uID := TRAY_ICON_ID;
  tnd.uCallbackMessage := WM_TRAY_CALLBACK;
  tnd.Wnd := FHandle;

  { Initialize properties to default values. }
  CurrentIcon := TIcon.Create;
  AlreadyShowingForm := false;

  FActive := false;
  FAutoPopup := false;
  FHint := EmptyStr;
  FIcon := TIcon.Create;
  FDisabledIcon := TIcon.Create;
  FState := tsEnabled;
  FShowHint := true;
  FOptions := [toMinimizeToTray, toUnclutterDelphi1TaskBar,toCloseFormsOnHideApp];

  { Initialize the Animation property. }
  FAnimation := TPGTrayIconAnimation.Create;
  FAnimation.FInterval := Animation_Interval;
  FAnimation.FIcons := TStringList.Create;
  FAnimation.FOptions := aoIconsAreFiles;
  FAnimation.TrayIcon := Self;
  FAnimation.AnimationTimer := TimerManager.AddTimer(false, FAnimation.FInterval, FAnimation.AnimationEvent);

  { Set the Icon's OnChange event to notify us when the tray icon's picture }
  { has been modified.  If the tray icon is visible when this property is }
  { modified, this forces the icon to update its display in the system tray. }
  FIcon.OnChange := IconChanged;
  FDisabledIcon.OnChange := IconChanged;

  { Replace the Applications's Minimize and Restore events with our own. }
  Application.OnMinimize := ApplicationMinimize;
  Application.OnRestore := ApplicationRestore;
end;

destructor TPGTrayIcon95.Destroy;
begin
  TimerManager.Enabled := false;
  TimerManager.Free;
  HideIcon;
  FIcon.Free;
  FDisabledIcon.Free;
  CurrentIcon.Free;
  DeallocateHWnd(FHandle);
  MessageWindow.Free;

  FAnimation.FIcons.Free;
  FAnimation.Free;

  inherited Destroy;
end;

procedure TPGTrayIcon95.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (AComponent = FPopupMenu) and (Operation = opRemove) then
    FPopupMenu := nil;
end;

{ Used to initially show the TrayIcon if it is supposed to be Active }
{ on Form Load. }
procedure TPGTrayIcon95.Loaded;
begin
  inherited Loaded;

  { Change the Application window so only it's icon displays in the TaskBar. }
  SetOptions(FOptions);

  if FActive and not (csDesigning in ComponentState) then
    begin
    FActive := false;
    ShowIcon;
    end;
end;

{ Minimizes the application to the system tray. }
procedure TPGTrayIcon95.ApplicationMinimize(Sender: TObject);
begin
  if csDesigning in ComponentState then
    exit;

  if FActive then
    begin
    if (toMinimizeToTray in FOptions) then
      HideApplication;
    end;

  FireEvent(FOnAppMinimize);
end;

{ Restores the application from the system tray. }
procedure TPGTrayIcon95.ApplicationRestore(Sender: TObject);
begin
  if csDesigning in ComponentState then
    exit;

  if FActive then
    ShowApplication;

  FireEvent(FOnAppRestore);
end;

procedure TPGTrayIcon95.ShowApplication;
begin
  PostEvent(DoShowApplication);
end;

procedure TPGTrayIcon95.DoShowApplication(Sender: TObject);
var
  AForm: TForm;
begin
  LockTaskBarUpdates(True);

  PGSetForegroundWindow(Application.Handle);
  
  { Restore the application if it isn't already restored. }
  if IsIconic(Application.Handle) then
    Application.Restore;

  AForm := Application.MainForm;
  if AForm = nil then
    begin
    WinProcs.SetParent(Application.Handle, 0);
    LockScreenUpdates(False);
    exit;
    end;

  try
    if not AForm.Visible then
      AForm.Show;
    if toHideTaskBarIconOnShowApp in FOptions then
      begin
      ShowWindow(Application.Handle, sw_Hide);
      ShowWindow(AForm.Handle, sw_Normal);
      NotifyOnFormClose(AForm);
      end
    else
      ShowWindow(Application.Handle, sw_Normal);
    WinProcs.SetParent(Application.Handle, 0);
  finally
    LockScreenUpdates(False);
  end;
end;

procedure TPGTrayIcon95.HideApplication;
var
  i: Integer;
begin
  if toCloseFormsOnHideApp in FOptions then
    begin
    LockScreenUpdates(True);
    WinProcs.SetParent(Application.Handle, PGApplicationHandle);
    for i := Screen.FormCount - 1 downto 0 do
      begin
      if Screen.Forms[i].Visible then
        CloseForm(Screen.Forms[i]);
      end;
    DoHideAppTaskBarIcon(Self);
    WinProcs.SetParent(Application.Handle, PGApplicationHandle);
    LockScreenUpdates(False);
    exit;
    end;

  PostEvent(DoHideApplication);
end;

procedure TPGTrayIcon95.DoHideApplication(Sender: TObject);
begin
  LockScreenUpdates(True);
  SetParent(Application.Handle, PGApplicationHandle);
  Application.NormalizeTopMosts;
  SetActiveWindow(Application.Handle);
  PGShowWinNoAnimate(Application.Handle, SW_MINIMIZE);
  if Assigned(FOnAppMinimize) then FOnAppMinimize(Self);
  ShowWindow(Application.Handle, sw_Hide);
  if Application.MainForm <> nil then
    ShowWindow(Application.MainForm.Handle, sw_Hide);
  LockScreenUpdates(False);
  DoHideAppTaskBarIcon(Self);
end;

function TPGTrayIcon95.GetClickedMouseButton: TMouseButton;
begin
  Result := MessageWindow.ClickedMouseButton;
end;

function TPGTrayIcon95.GetHandle: THandle;
begin
  Result := FHandle;
end;

function TPGTrayIcon95.GetHint: string;
begin
  Result := StrPas(tnd.szTip);
end;

procedure TPGTrayIcon95.SetHint(Value: string);
begin
  if FHint = Value then
    exit;
  FHint := Value;

  { Update the hint if it is currently being displayed. }
  if FShowHint then
    SetShowHint(FShowHint);
end;

procedure TPGTrayIcon95.SetIcon(Value: TIcon);
begin
  FIcon.Assign(Value);
  LoadCurrentIcon;
end;

procedure TPGTrayIcon95.SetDisabledIcon(Value: TIcon);
begin
  FDisabledIcon.Assign(Value);
  LoadCurrentIcon;
end;

{ Show/Hide the icon depending on its current value. }
procedure TPGTrayIcon95.SetActive(AnActive: Boolean);
begin
  { Dont bother showing the icon if it's design time. }
  if (csDesigning in ComponentState) then
    begin
    FActive := AnActive;
    exit;
    end;

  if AnActive then
    ShowIcon
  else
    HideIcon;
end;

{ Send a message to Windows to inform it that we would like to }
{ add/remove/update our tray icon. }
procedure TPGTrayIcon95.NotifyShell(dwMessage: LongInt);
{$IFDEF WIN32}
var
  tnd32: TNOTIFYICONDATA;
{$ENDIF}
begin
  if (csReading in ComponentState) then
    exit;

{$IFDEF WIN32}
  { Delphi 2.0 Specific Code. }
  tnd32.cbSize := SizeOf(TNOTIFYICONDATA);
  tnd32.wnd := tnd.wnd;
  tnd32.uID := tnd.uID;
  tnd32.uFlags := tnd.uFlags;
  tnd32.hIcon := CurrentIcon.Handle;
  StrPCopy(tnd32.szTip, tnd.szTip);
  tnd32.uCallbackMessage := tnd.uCallbackMessage;
  Shell_NotifyIcon(dwMessage, @tnd32);
{$ELSE}
  { Delphi 1.0 Specific Code. }
  tnd.cbSize  := SizeOf(PGNOTIFYICONDATA);
  tnd.hIcon := HIcon32;

  Shell_NotifyIcon(dwMessage, @tnd, idShell_NotifyIcon);
{$ENDIF}
end;

{ Used to delay firing an event.  Sent from Self (ie PGTrayIcon95) }
procedure TPGTrayIcon95.PostEvent(AnEvent: TNotifyEvent);
begin
  PostEventFromObject(AnEvent, Self);
end;

{ Used to delay firing an event. Sent from any object. }
procedure TPGTrayIcon95.PostEventFromObject(AnEvent: TNotifyEvent; ASender: TObject);
begin
  PostedEvent := AnEvent;
  PostedEventSender := ASender;
  WinProcs.PostMessage(FHandle, WM_PGPOSTEVENT, 0, 0);
end;

{ Add the icon to the system tray.}
procedure TPGTrayIcon95.ShowIcon;
begin
  if FActive then
    exit;

  { If the animation is active, then use the first icon in the animation }
  { for the icon initially displayed in the tray. Otherwise, use the }
  { default static icon. }
  LoadCurrentIcon;

  { Add the icon to the tray. }
  FActive := true;
  tnd.uFlags := PGNIF_MESSAGE or PGNIF_ICON or PGNIF_TIP;
  NotifyShell(PGNIM_ADD);

  { If the icon should be animated, start the animation. }
  if FState = tsAnimated then
    FAnimation.StartAnimation;
end;

{ Remove the icon from the system tray.}
procedure TPGTrayIcon95.HideIcon;
begin
  if not FActive then
    exit;
  FActive := false;
  if FState = tsAnimated then
    FAnimation.StopAnimation;

  NotifyShell(PGNIM_DELETE);
end;

procedure TPGTrayIcon95.LockScreenUpdates(ShouldLock: Boolean);
const
  RefCount: Integer = 0;
begin
  if ShouldLock then
    Inc(RefCount)
  else
    Dec(RefCount);

  if (RefCount = 1) then
    LockWindowUpdate(GetDesktopWindow)
  else if RefCount = 0 then
    LockWindowUpdate(0);
end;

{ Shows a form modally, and attempts to stop the Applications taskbar }
{ icon from showing. }
function TPGTrayIcon95.ShowModalForm(const AForm: TForm): Integer;
begin
  Result := DoShowForm(AForm, True);
end;

{ Shows a form, and attempts to stop the Applications taskbar icon }
{ from showing. }
procedure TPGTrayIcon95.ShowForm(const AForm: TForm);
begin
  DoShowForm(AForm, False);
end;

{ Closes a form and attempts to stop the Applications taskbar icon }
{ from showing.  You should never need to call this function }
{ explicitely, since it is implicitely called by the TrayIcon when }
{ forms that were shown using the ShowModalForm and ShowForm }
{ are closed. }
procedure TPGTrayIcon95.CloseForm(const AForm: TForm);
var
  ShouldHideApplication: Boolean;
begin
  if AForm = nil then
    begin
    raise EInvalidOperation.Create('You cant hide a "nil" form !');
    exit;
    end;

  ShouldHideApplication := ApplicationIsHidden;
  LockScreenUpdates(True);
  try
    if ShouldHideApplication then
      begin
      WinProcs.SetParent(Application.Handle, PGApplicationHandle);
      ShowWindow(Application.Handle, sw_Hide);
      end;
    if AForm = Application.MainForm then
      AForm.Hide
    else
      AForm.Close;
  finally
    LockScreenUpdates(False);
    if ShouldHideApplication or (AForm = Application.MainForm) then
      PostEventFromObject(DoHideAppTaskBarIcon, AForm);
  end;
end;

{ Hides the Applications' TaskBar icon. }
procedure TPGTrayIcon95.DoHideAppTaskBarIcon(Sender: TObject);
begin
  LockScreenUpdates(True);
  try
    WinProcs.SetParent(Application.Handle, PGApplicationHandle);
    ShowWindow(Application.Handle, sw_Hide);
    WinProcs.SetParent(Application.Handle, 0);
  finally
    LockScreenUpdates(False);
  end;
end;

procedure TPGTrayIcon95.LockTaskBarUpdates(ShouldLock: Boolean);
begin
  if ShouldLock then
    begin
    LockScreenUpdates(True);
    WinProcs.SetParent(Application.Handle, PGApplicationHandle);
    end
  else
    begin
    ShowWindow(Application.Handle, sw_Hide);
    WinProcs.SetParent(Application.Handle, 0);
    LockScreenUpdates(False);
    end;
end;

procedure TPGTrayIcon95.DoSetForegroundWindow(Sender: TObject);
var
  AForm: TForm;
begin
  LockScreenUpdates(True);
  DoHideAppTaskBarIcon(Self);
  AForm := TForm(Sender);
  PGSetForegroundWindow(AForm.Handle);
  LockScreenUpdates(False);
end;

procedure TPGTrayIcon95.DoFormOnShow(Sender: TObject);
var
  AForm: TForm;
begin
  AForm := TForm(Sender);
  try
    if Assigned(SaveOnShowEvent) then
      SaveOnShowEvent(AForm);
    LockScreenUpdates(False);
    AForm.OnShow := SaveOnShowEvent;
  finally
    if ApplicationWasHidden then
      PostEventFromObject(DoSetForegroundWindow, AForm);
    AlreadyShowingForm := false;
  end;
end;

function TPGTrayIcon95.ApplicationIsHidden: Boolean;
begin
  Result := not IsWindowVisible(Application.Handle);
end;


{ This function attempts to implement showing a form without having it's }
{ associated taskbar icon display when shown.  This problem arose in Delphi 3 }
{ due to some changes Borland made to the Forms.pas file. }
{ See procedure TApplication.UpdateVisible for more details.  It is called whenever }
{ a form is created, destroyed, hidden or shown, and causes the Application Window }
{ to reappear in the TaskBar. }
function TPGTrayIcon95.DoShowForm(const AForm: TForm; const ShowModal: Boolean): Integer;
begin
  Result := 0;
  if AForm = nil then
    begin
    raise EInvalidOperation.Create('You cant show a "nil" form !');
    exit;
    end;

  ApplicationWasHidden := ApplicationIsHidden;
  if IsIconic(Application.Handle) then
    Application.Restore;

  PGSetForegroundWindow(Application.Handle);
      
  if AForm.Visible or IsWindowVisible(AForm.Handle) then
    begin
    DoSetForegroundWindow(AForm);
    exit;
    end;

  if AlreadyShowingForm then
    begin
    raise EInvalidOperation.Create('You cant call ShowForm or ShowModalForm in the OnShow event of a form !');
    exit;
    end;

  AlreadyShowingForm := true;
  try
    { Replace the OnShow event of the Form with our own OnShow. }
    { Once our OnShow gets called, we'll allow the form that was }
    { shown to be displayed, and put back any OnShow event that was }
    { previously there. }
    SaveOnShowEvent := AForm.OnShow;
    AForm.OnShow := DoFormOnShow;
    NotifyOnFormClose(AForm);

    { Show the form. }
    LockScreenUpdates(True);
    if ApplicationWasHidden then
      WinProcs.SetParent(Application.Handle, PGApplicationHandle);

    if ShowModal then
      Result := AForm.ShowModal
    else
      AForm.Show;
  finally
    AlreadyShowingForm := false;
  end;
end;

{ Used implicitely to hide a form without showing the Applications TaskBar icon. } 
function TPGTrayIcon95.NotifyOnFormClose(AForm: TForm): Boolean;
var
  ANotification: TPGNotifyOnFormClose;
  ACloseEvent: TNotifyEvent;
begin
  Result := Assigned(AForm);
  if not Result then
    exit;

  ANotification := TPGNotifyOnFormClose.Create(AForm);
  ANotification.AddCloseEvent(AForm, FormOnCloseNotification);
  ACloseEvent := DoHideAppTaskBarIcon;
end;

procedure TPGTrayIcon95.FormOnCloseNotification(Sender: TObject);
var
  ShouldHideApplication: Boolean;
  AForm: TForm;
begin
  AForm := TForm(Sender);
  ShouldHideApplication := ApplicationIsHidden;
  LockScreenUpdates(True);
  try
    if ShouldHideApplication then
      begin
      WinProcs.SetParent(Application.Handle, PGApplicationHandle);
      ShowWindow(Application.Handle, sw_Hide);
      end;
  finally
    LockScreenUpdates(False);
    if ShouldHideApplication then
      PostEventFromObject(DoHideAppTaskBarIcon, AForm);
  end;
end;

{ Modify some information about the icon that is in the system tray. }
procedure TPGTrayIcon95.Refresh;
begin
  tnd.uFlags := PGNIF_MESSAGE or PGNIF_ICON or PGNIF_TIP;
  NotifyShell(PGNIM_MODIFY);
end;

procedure TPGTrayIcon95.RefreshTaskList;
var
  i: Integer;
begin
  if ApplicationIsHidden and not IsIconic(Application.Handle) then
    begin
    for i := 0 to Screen.FormCount - 1 do
      begin
      if Screen.Forms[i].Visible then
        exit;
      end;
    end;

  PGSetForegroundWindow(GetDesktopWindow);
end;

{ Returns whether or not we should fire the OnMouseMove event.  Note that }
{ it isn't fired if an event isn't defined, or if we're currently processing }
{ a click/double click message. }
function TPGTrayIcon95.ShouldDoMouseMove: Boolean;
begin
  Result := Assigned(FOnMouseMove);
  Result := Result and not MessageWindow.WaitForClickEventTimer.Enabled;
  Result := Result and MessageWindow.AlreadyFiredMouseEnterEvent;
end;

{ Returns whether or not we should fire the OnMouseEnter/OnMouseLeave events. }
function TPGTrayIcon95.ShouldDoMouseEnterOrMouseExit: Boolean;
begin
  Result := Assigned(FOnMouseEnter) or Assigned(FOnMouseExit);
end;

{ Returns whether or not the tray icon should display a popup menu }
{ when the tray icon is "right clicked". }
function TPGTrayIcon95.ShouldDisplayPopup: Boolean;
begin
  Result := FAutoPopup and (FPopupMenu <> nil);
end;

{ Displays the popup menu at the current mouse position. }
procedure TPGTrayIcon95.ShowPopupAtCursor;
begin
  if not Assigned(FPopupMenu) then
    exit;
    
  { When you display a context menu for a TrayIcon, clicking anywhere besides }
  { the menu or the window that created the menu (if it is visible) doesn't }
  { cause the menu to disappear. When this behavior is corrected, the }
  { second time this menu is displayed, it displays and then immediately disappears. }
  { To correct the first behavior, you need to make the current window the }
  { foreground window before calling PopupMenu.Popup(x,y) }
  PGSetForegroundWindow(Application.Handle);
  Application.ProcessMessages;
  PopupMenu.Popup(MessageWindow.ptMousePosition.X, MessageWindow.ptMousePosition.Y);

  { The second problem is caused by a problem with TrackPopupMenu. It is necessary }
  { to force a task switch to the application that called TrackPopupMenu at some }
  { time in the near future. This can be accomplished by posting a benign message to }
  { the window or thread. }
  PostMessage(Application.Handle, WM_USER, 0, 0);
  RefreshTaskList;
end;

procedure TPGTrayIcon95.SetShowHint(Value: Boolean);
var
  LengthOfTip: Byte;
  NewHint: string;

  function Min(A,B :Byte): Byte;
  begin
    if A < B then
      Result := A
    else
      Result := B;
  end;

begin
  FShowHint := Value;

  { If the hint should be displayed, copy the hint string to the }
  { notification data structure, otherwise, copy an empty string }
  { to the data structure so the hint won't be displayed. }
  if FShowHint then
    begin
    LengthOfTip := Min(NID_MAXTIP, Length(FHint));
    NewHint := Copy(FHint, 1, LengthOfTip);
    end
  else
    NewHint := EmptyStr;

  { Notify the icon that it needs to update the display state }
  { of the hint. (i.e. To hide or show the hint as appropriate.) }
  StrPCopy(tnd.szTip, NewHint);
  if FActive and not (csDesigning in ComponentState) then
    Refresh;
end;

procedure TPGTrayIcon95.SetOptions(Value: TPGTrayIconOptions);
{$IFNDEF WIN32}
var
  Style: LongInt;
{$ENDIF}
begin
  FOptions := Value;

{$IFNDEF WIN32}
  { Remove the extra task bar icons for the task bar if the user wants to get }
  { rid of them.  Only relevant to applications written in Delphi 1.0 }
  if (toUnclutterDelphi1TaskBar in FOptions) and not (csDesigning in ComponentState) then
    begin
    Style := GetWindowLong(Application.Handle, GWL_STYLE);
    Style := Style or WS_CAPTION;
    SetWindowLong(Application.Handle, GWL_STYLE, Style);
    end;
{$ENDIF}
end;

procedure TPGTrayIcon95.SetState(AState: TPGTrayIconState);
const
  StateChanging: Boolean = false;
begin
  { If someone is trying to change the state property in the OnStateChanging }
  { event, let them know that they should modify the "NewState" parameter, }
  { and not the State property itself.  This ensures that an infinite loop of }
  { state change events never occurs. }
  if StateChanging then
    begin
    StateChanging := false;
    raise EInvalidOperation.Create('You cannot modify the state property in the OnStateChanging event !!!');
    end;

  if (not FActive) or (csReading in ComponentState) then
    begin
    FState := AState;
    exit;
    end;

  if FState = AState then
    exit;

  { Find out if the change of state should be cancelled. }
  StateChanging := true;
  if Assigned(FOnStateChanging) then
    FOnStateChanging(Self, AState);
  StateChanging := false;

  { Exit if the change of state was cancelled. A change of state can be }
  { cancelled in the OnStateChanging by setting the "NewState" }
  { parameter to the current state value. }
  if FState = AState then
    exit;

  { If the tray icon is currently being animated, then stop the animation. }
  if FState = tsAnimated then
    FAnimation.StopAnimation;

  { Change the state of the tray icon to its new state. }
  FState := AState;

  { Show the icon in the tray. }
  LoadCurrentIcon;

  { Animate the tray icon if it needs animation. }
  if FState = tsAnimated then
    FAnimation.StartAnimation;
end;

{ Fire's an event if it has been defined. }
procedure TPGTrayIcon95.FireEvent(AnEvent:TNotifyEvent);
begin
  if Assigned(AnEvent) then
    AnEvent(Self);
end;

{ Update's the icon's display if it is currently being shown }
{ in the system tray. }
procedure TPGTrayIcon95.IconChanged(Sender: TObject);
begin
  { Dont bother showing the icon if it's design time or the icon is nil. }
  if (csDesigning in ComponentState) then
    exit;

  { Update the icon's display. }
  if FActive then
    LoadCurrentIcon;
end;

{ Loads the animation icon corresponding to "AnIconIndex" }
procedure TPGTrayIcon95.LoadAnimationIcon(const AnIconIndex: Integer);
var
  IconFilename: array[0..255] of char;
begin
  if AnIconIndex >= FAnimation.FIcons.Count then
    exit;

  try
    if FAnimation.FOptions = aoIconsAreFiles then
      { Load the animation icon from an icon file. }
      CurrentIcon.LoadFromFile(FAnimation.FIcons[AnIconIndex])
    else
      begin
      { Load the animation icon from an icon resource }
      { contained in the current executable. }
      StrPCopy(IconFilename, UpperCase(FAnimation.FIcons[AnIconIndex]));

      { Load the selected icon into the current icon.  Note that this }
      { is being done differently under Delphi 2+ to solve a drawing }
      { problem (black marks) with the icon under WinNT 4.X }
{$IFDEF WIN32}
      CurrentIcon.Handle := WinProcs.LoadImage(HInstance,
          IconFilename, IMAGE_ICON, 16, 16, LR_DEFAULTCOLOR);
{$ELSE}
      CurrentIcon.Handle := WinProcs.LoadIcon(HInstance, IconFilename);
{$ENDIF}

      end;
  except
    on EFOpenError do;
    else
      raise;
  end;
end;

{ Loads the "CurrentIcon" with whichever icon should be displayed in the }
{ system tray, depending on the current state of the icon. }
procedure TPGTrayIcon95.LoadCurrentIcon;
begin
  if csDesigning in ComponentState then
    exit;

  { Animate, Enable, or Disable the icon, depending on it's current state. }
  case FState of
    tsAnimated: LoadAnimationIcon(FAnimation.FIconIndex);
    tsEnabled:  CurrentIcon.Assign(FIcon);
    tsDisabled: CurrentIcon.Assign(FDisabledIcon);
  end;

  { Update the icon that is being displayed in the system tray. }
  if FActive then
    Refresh;
end;

{ This routine is used to get a 32 bit HIcon for the current icon }
{ being displayed in the system tray. }
function TPGTrayIcon95.HIcon32: LongInt;
{$IFNDEF WIN32}
var
  PGIconWindowHandle32: LongInt;
{$ENDIF}
const
  PGWOW_FULLHWND: LongInt = 14;
  PGGCL_HICON: LongInt = -14;
begin
{$IFDEF WIN32}
  { If we're using Delphi 2.0, just return the handle, since it's already }
  { a 32 bit icon handle. }
  Result := CurrentIcon.Handle;
{$ELSE}
  { If we're using Delphi 1.0 and running under Windows NT we have }
  { to convert the 16 bit HIcon to a 32 bit HIcon.  Since the Generic }
  { thunking mechanism doesn't directly support converting HIcon's to 32 }
  { bits, we have to improvise.  By setting the HIcon of our own 16 bit Window}
  { class to the HIcon of the CurrentIcon under Win16, and requesting the }
  { HIcon of the Window Class under Win32, we can get Windows to do the }
  { conversion for us implicitely. }
  if RunningWinNT then
    begin
    { Set the HIcon of the class to the HIcon that we want to convert. }
    SetClassWord(PGApplicationHandle, GCW_HICON, CurrentIcon.Handle);

    { Ask Win32 for the HIicon of the class by using the Win32 equivalent }
    { of Win31's GetClassWord, Win32's GetClassLong. }
    PGIconWindowHandle32 := PGWowHandle32(PGApplicationHandle, PGWOW_FULLHWND, idPGWowHandle32);
    Result := PGGetClassLong(PGIconWindowHandle32, PGGCL_HICON, idPGGetClassLong);
    end
  else
    { If we're using Delphi 1.0 and running Windows 95, just return the }
    { handle since it's already equivalent to a 32 bit icon handle. }
    Result := CurrentIcon.Handle;
{$ENDIF}
end;

{ TPGTimerItem }
constructor TPGTimerItem.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  TimerManager := TPGTimerManager(AOwner);
end;


{ TPGTimerManager - Manages a group of timed events using a single Windows Timer. }
constructor TPGTimerManager.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Timer := TTimer.Create(Self);
  FEnabled := false;

  Timer.Enabled := false;
  Timer.OnTimer := DoTimeouts;
  
  TimerItems := TList.Create;
end;

destructor TPGTimerManager.Destroy;
begin
  Timer.Enabled := false;
  Timer.Free;
  TimerItems.Free;
  inherited Destroy;
end;

{ Add a new timer to the list of currently defined timers. }
function TPGTimerManager.AddTimer(AnEnabled: Boolean; AnInterval: TPGTimerInterval; AnEvent: TNotifyEvent): TPGTimerItem;
var
  ATimerItem: TPGTimerItem;
begin
  ATimerItem := TPGTimerItem.Create(Self);
  ATimerItem.Enabled := AnEnabled;
  ATimerItem.Interval := AnInterval;
  ATimerItem.OnTimer := AnEvent;
  TimerItems.Add(ATimerItem);
  
  Result := ATimerItem;
  if ATimerItem.Enabled then
    RefreshInterval;
end;

{ Removes a timer from the list of timers. }
procedure TPGTimerManager.DeleteTimer(AnId: Integer);
begin
  TimerItems.Delete(AnId);
  RefreshInterval;	
end;

procedure TPGTimerManager.SetEnabled(Value: Boolean);
begin
  if FEnabled = Value then
    exit;
    
  FEnabled := Value;
    
  { Note that RefreshInterval takes care of enabling or disabling }
  { the real timer as is required. }
  if FEnabled then
    RefreshInterval;
end;

{ Refreshes the Main Timer Interval that is used to calculate and }
{ fire the Timer Event for all Timer Items.  Note that the overall }
{ interval is calculated to be the smallest Timer Item interval }
{ or 150, whichever is smaller. }
procedure TPGTimerManager.RefreshInterval;
var
  i: Integer;
  AnInterval: TPGTimerInterval;
  BestInterval: LongInt;
  ATimerItem: TPGTimerItem;
  SomeTimersAreEnabled: Boolean;
begin
  if not FEnabled then
    exit;

  { Turn off the timer while we're refreshing it's interval. }
  Timer.Enabled := false;

  { Make the count interval will always be within 150 ms of }
  { any of the other suggested timer values. }
  BestInterval := 150;

  { Now find out the smallest interval currently enabled. }
  SomeTimersAreEnabled := false;
  for i := 0 to TimerItems.Count - 1 do
    begin
    ATimerItem := Timers(i);
    if not ATimerItem.Enabled and Assigned(ATimerItem.OnTimer) then
      continue;
    SomeTimersAreEnabled := true;

    AnInterval := ATimerItem.Interval;
    if AnInterval < BestInterval then
      BestInterval := AnInterval;

    { Make sure the timer remembers it's starting tick count. }
    ATimerItem.LastTickCount := GetTickCount;
    end;

  { Update the timer interval and restart the timer if it was going. }
  if FEnabled and SomeTimersAreEnabled then
    begin
    Timer.Interval := BestInterval;
    Timer.Enabled := True;
    end
  else
    Timer.Enabled := False;
end;

function TPGTimerManager.Timers(AnIndex: Integer): TPGTimerItem;
begin
  Result := TPGTimerItem(TimerItems[AnIndex]);
end;

function TPGTimerManager.TimerCount: Integer;
begin
  Result := TimerItems.Count;
end;

procedure TPGTimerManager.DoTimeouts(Sender: TObject);
var
  i: Integer;
  ATimerItem: TPGTimerItem;
begin
  for i := 0 to TimerItems.Count - 1 do
    begin
    ATimerItem := Timers(i);
    if ATimerItem.Enabled and Assigned(ATimerItem.OnTimer) then
      begin
      if ATimerItem.ShouldTimeout then
        ATimerItem.Timeout;
      end;
    end;
end;

{ Returns whether or not a timer item should "timeout", and fire it's }
{ OnTimer event.  Note that the time between timeout's is calculated by }
{ using the Windows system tick count, not the time interval of the }
{ actual windows timer.  This provides much more accurate and "timely" response }
{ for all events in the timer pool. } 
function TPGTimerItem.ShouldTimeout: Boolean;
begin
  Result := ((GetTickCount - LastTickCount) >= Interval);
end;

procedure TPGTimerItem.TimeOut;
begin
  OnTimer(Self);
  LastTickCount := GetTickCount;
end;

procedure TPGTimerItem.SetEnabled(Value: Boolean);
begin
  if FEnabled = Value then
    exit;
  FEnabled := Value;
  TimerManager.RefreshInterval;
end;

procedure TPGTimerItem.SetInterval(Value: TPGTimerInterval);
begin
  if FInterval = Value then
    exit;
  FInterval := Value;
  if FEnabled then
    TimerManager.RefreshInterval;
end;


{ TPGNotifyOnFormClose }
procedure TPGNotifyOnFormClose.DoFormOnClose(Sender: TObject; var Action: TCloseAction);
begin
  if Assigned(NotifyOnFormCloseEvent) then
    NotifyOnFormCloseEvent(Form);
  if Assigned(FormOnClose) then
    FormOnClose(Sender, Action);
end;

procedure TPGNotifyOnFormClose.AddCloseEvent(AForm: TForm; AnEvent: TNotifyEvent);
var
  i: Integer;
  AVcl: TComponent;
begin
  for i := 0 to AForm.ComponentCount - 1 do
    begin
    AVcl := AForm.Components[i];
    if (AVcl is TPGNotifyOnFormClose) and not (AVcl = Self) then
      begin
      Free;
      exit;
      end;
    end;

  Form := AForm;
  FormOnClose := AForm.OnClose;
  NotifyOnFormCloseEvent := AnEvent;

  AForm.OnClose := DoFormOnClose;
end;

destructor TPGNotifyOnFormClose.Destroy;
begin
  inherited Destroy;
end;


{ Required for the Delphi 1.0 Solution. }
{$IFNDEF WIN32}
procedure DoneTrayApplication; far;
begin
  with Application do
  begin
    if Handle <> 0 then ShowOwnedPopups(Handle, False);
    Destroying;
    DestroyComponents;
  end;
end;
{$ENDIF}

procedure RunTrayApplication;
begin
{$IFDEF WIN32}
  { Required for the Delphi 2.0 Solution. }
  Application.ShowMainForm := false;
  Application.Run;
{$ELSE}
  { Required for the Delphi 1.0 Solution. }
  ShowWindow(Application.Handle, sw_Hide);
  AddExitProc(DoneTrayApplication);
  repeat
    Application.HandleMessage;
  until Application.Terminated;
{$ENDIF}
end;

initialization
  { Initialize the time to wait between a single click and a double click. }
  { Note that the default double click time as defined by Windows is 500ms, }
  { but seems too slow, so we're setting a slightly lower default value here. }
  PGTrayIconDoubleClickTime := 400;

  { Delphi 1.0 specific code}
{$IFNDEF WIN32}
  { Initialize Generic Thunking Interface for calling Win32 functions from }
  { 16 bit Delphi 1.0 applications. }
  @Callproc32W:=nil;
  @FreeLibrary32W:=nil;
  @GetProcAddress32W:=nil;
  @LoadLibraryEx32W:=nil;
  @lpvGetLastError:=nil;
  lpvGetCapture:=nil;
  exitsave:=exitproc;
  exitproc:=@cleanuplibs;

  { Each function must be declared with Declare32. The handle returned }
  { by Declare32 must be passed as the last parameter of the function when }
  { the function is called. }

  { Parameters of Declare32: }
  {   First:  The name of the original win32 function: CASE SENSITIVE!!!!! }
  {   Second: The name of the 32 bit module where the function is located. }
  {   Third:  A string describing all parameters. }
  {           ( p=pointer, i=longint, w=Windows handle )    }
  @Shell_NotifyIcon  := @Call32;
  idShell_NotifyIcon := Declare32('Shell_NotifyIcon', 'shell32', 'ip');

  @PGWowHandle32  := @Call32;
  idPGWowHandle32 := Declare32('WOWHandle32','WOW32','ii');

  @PGGetClassLong  := @Call32;
  idPGGetClassLong := Declare32('GetClassLong','USER32.DLL','ii');

  @SetForegroundWindow  :=@Call32;
  idSetForegroundWindow := Declare32('SetForegroundWindow', 'user32', 'w');
{$ENDIF}

{$IFDEF PGShareware_Version}
  { Display shareware registration information when the user is running }
  { an executable outside of Delphi. }
  if not PGDelphiIsRunning then
    PGShowSharewareNotice('TPGTRAYICON95');
{$ENDIF}
end.
